from typing import Any
from abc import ABC, abstractmethod

import json
from pika import BlockingConnection

from .types import Subjects

class AbstractPublisher(ABC):
    @property
    @abstractmethod
    def subject(self) -> Subjects | str:
        raise NotImplementedError()

    @abstractmethod
    def publish(self, data: Any):
        pass

    def on_publish(self, data: Any):
        self.channel.basic_publish(exchange='', routing_key=self.subject, body=json.dumps(data).encode('utf-8'))

    def __init__(self, connection: BlockingConnection):
        self.channel = connection.channel()  # Start a channel.
        self.channel.queue_declare(queue=self.subject)  # Declare a queue.



from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession

from ..config import settings


def create_async_session():
    async_engine = create_async_engine(settings.ASYNC_SQLALCHEMY_DATABASE_URI)

    session = AsyncSession(async_engine)
    session.sync_session.expire_on_commit = False

    return session


async def get_async_session():
    async with create_async_session() as session:
        yield session



from typing import Any
from abc import ABC, abstractmethod

import json
from pika import BlockingConnection

from .types import Subjects


class AbstractConsumer(ABC):
    @property
    @abstractmethod
    def subject(self) -> Subjects | str:
        raise NotImplementedError()

    def callback(self, ch, method, properties, body: bytes):
        self.on_message(json.loads(body.decode('utf-8')))

    @abstractmethod
    def on_message(self, data: Any):
        raise NotImplementedError()

    def __init__(self, connection: BlockingConnection):
        self.channel = connection.channel()  # Start a channel.
        self.channel.queue_declare(queue=self.subject)  # Declare a queue.
        self.channel.basic_consume(queue=self.subject, on_message_callback=self.callback, auto_ack=True)

    def run(self):
        print(f"Start consuming on queue '{self.subject}'...")
        self.channel.start_consuming()



from typing import TypedDict
from enum import Enum


class Subjects(Enum):
    TargetCreated = 'target.created'
    StructureComputed = 'structure.computed'


class TargetCreatedData(TypedDict):
    id: int
    custom: bool
    rna_id: str
    spliced: bool
    seq: str


class StructureComputedData(TypedDict):
    target_id: int
    id: int
    openness: list[float]



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ..config import settings

if os.environ.get('FASTAPI_ENV') == 'dev_inmemory' or os.environ.get('FASTAPI_ENV') == 'prod_inmemory':
    engine = create_engine(settings.SQLALCHEMY_DATABASE_URI, connect_args={"check_same_thread": False})
else:
    engine = create_engine(settings.SQLALCHEMY_DATABASE_URI)


def create_session():
    session = Session(autocommit=False, autoflush=False, bind=engine, expire_on_commit=False)

    return session


def get_session():
    session = Session(autocommit=False, autoflush=False, bind=engine, expire_on_commit=False)

    try:
        yield session
    finally:
        session.close()



from typing import Any, Optional
from enum import Enum
import operator

import strawberry

operator_keys = {
    '': operator.eq,
    '>': operator.ge,
    '<': operator.le,
}


def split(item_list: Any, n: int):
    k, m = divmod(len(item_list), n)

    return (item_list[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))


@strawberry.enum(description='Selection of organisms.')
class OrganismSelect(Enum):
    NONE = ''
    HUMAN = 'human'
    MOUSE = 'mouse'
    RAT = 'rat'


organism_mapping = {
    'human': 9606,
    'mouse': 10090,
    'rat': 10116,
}


@strawberry.input
class InputWithOrganism:
    organism: Optional[OrganismSelect] = OrganismSelect.NONE



from datetime import datetime
from math import ceil
from typing import Any, Dict, Iterator, List, Optional, Tuple
from dataclasses import dataclass

import strawberry
from sqlalchemy import Column, DateTime, func, Integer, String, Table
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import Query

from .db import create_session, engine


@dataclass(init=False)
class CreateUpdateFields:
    # Keep track when records are created and updated.
    created_at: Optional[datetime] = Column(DateTime(), index=True, default=datetime.utcnow)
    updated_at: Optional[datetime] = Column(DateTime(), index=True, default=datetime.utcnow, onupdate=datetime.utcnow)
    created_by = Column(String)
    updated_by = Column(String)


class ResourceMixin(CreateUpdateFields):
    __table__: Table

    @classmethod
    def _bulk_insert(cls, data, label: str, dtype: str = '') -> None:
        """
        Bulk insert data to the model and log it. This is much more efficient than adding 1 row at a time in a loop.

        :param data: Data to be saved
        :type data: list
        :param dtype: Data type
        :type dtype: str
        :param label: Label for the output
        :type label: str
        :return: None
        """
        engine.execute(cls.__table__.insert(), data)
        print(f'Finished inserting {len(data):,} {(dtype + " ") if dtype else ""}{label}.')

        return None


class ResourceMixinWithVersion(ResourceMixin):
    version: Optional[int] = Column(Integer, nullable=False)
    __mapper_args__ = {'version_id_col': version}


@dataclass(init=False)
class RefSeqFields:
    id: strawberry.ID = Column(String(24), primary_key=True)  # RefSeq acc version, e.g. NM_003331.5 / NP_003322.3.
    acc: str = Column(String(16), nullable=False, unique=True)


class RefseqMixin(RefSeqFields):
    @classmethod
    def find_by_refseq_id(cls, refseq_id: str):
        """Find a model by its RefSeq accession ID."""
        with create_session() as db:
            return db.query(cls).filter((cls.id == refseq_id) | (cls.acc == refseq_id.split('.')[0])).first()


class ExternalResourceMixin:
    # Keep track when records are created and updated.
    created_at = Column(DateTime(), index=True, default=datetime.utcnow)
    updated_at = Column(DateTime(), index=True, default=datetime.utcnow, onupdate=datetime.utcnow)
    created_by = Column(String)
    updated_by = Column(String)

    # @classmethod
    # def get_by_id(cls, id: Union[int, str, List[str]]):
    #     try:
    #         return cls.query.get(id)
    #     except ValueError:
    #         return None


# def sort_query(model: Any, query: Query, sort_keys: Dict[str, InstrumentedAttribute],
#                order_by: Iterator[str]) -> Query:
#     """Sort list with order_by fields, append id_ASC/id_DESC if not present."""
#     sort_list = [order.split('_') for order in order_by]
#     query = query.order_by(*[sort_keys[sort_key].desc() if sort_order == 'DESC' else sort_keys[sort_key]
#                              for (sort_key, sort_order) in sort_list if sort_key in sort_keys])
#     if not ('id_ASC' in order_by or 'id_DESC' in order_by):
#         query = query.order_by(model.id.desc() if sort_list[0][1] == 'DESC' else model.id)
#
#     return query


def sort_query(model: Any, query: Query, sort_keys: Dict[str, Any], order_by: Iterator[str]) -> Query:
    """Sort list with order_by fields, append id_ASC/id_DESC if not present."""
    sort_list = [order.split('_') for order in order_by]
    # query = query.order_by(*[sort_keys[sort_key].desc() if sort_order == 'DESC' else sort_keys[sort_key]
    #                          for (sort_key, sort_order) in sort_list if sort_key in sort_keys])
    ordering_params = []
    for sort_key, sort_order in sort_list:
        if sort_key in sort_keys:
            if sort_order == 'DESC':
                if isinstance(sort_keys[sort_key], Tuple):
                    for ordering_param in sort_keys[sort_key]:
                        ordering_params.append(ordering_param.desc())
                else:
                    ordering_params.append(sort_keys[sort_key].desc())
            else:
                if isinstance(sort_keys[sort_key], Tuple):
                    for ordering_param in sort_keys[sort_key]:
                        ordering_params.append(ordering_param)
                else:
                    ordering_params.append(sort_keys[sort_key])

    query = query.order_by(*ordering_params)
    if not ('id_ASC' in order_by or 'id_DESC' in order_by):
        query = query.order_by(model.id.desc() if sort_list[0][1] == 'DESC' else model.id)

    return query


class Pagination(object):
    def __init__(self, items: List[Any], page: int, per_page: int, total: int):
        self.page = page
        self.items = items
        self.prev_page = None
        self.next_page = None
        self.has_prev = page > 1
        if self.has_prev:
            self.prev_page = page - 1
        previous_items = (page - 1) * per_page
        self.has_next = previous_items + len(items) < total
        if self.has_next:
            self.next_page = page + 1
        self.total = total
        self.pages = int(ceil(total / float(per_page)))


def paginate(query: Query, page: int, per_page: int):
    if page <= 0:
        raise AttributeError('page needs to be >= 1')
    if per_page <= 0:
        raise AttributeError('per_page needs to be >= 1')
    if per_page > 100:
        raise AttributeError('per_page needs to be <= 100')

    items = query.limit(per_page).offset((page - 1) * per_page).all()
    total = query.order_by(None).count()
    # total = session.execute(query.statement.with_only_columns([func.count()]).order_by(None)).scalar()

    return Pagination(items, page, per_page, total)


async def async_paginate(session: AsyncSession, query: Query, page: int, per_page: int):
    if page <= 0:
        raise AttributeError('page must be >= 1')
    if per_page <= 0:
        raise AttributeError('per_page must be >= 1')
    if per_page > 100:
        raise AttributeError('per_page needs to be <= 100')

    paginated_query = query.limit(per_page).offset((page - 1) * per_page)
    result = await session.execute(paginated_query)

    items = result.scalars().all()
    total_result = await session.execute(select(func.count()).select_from(query))
    total = total_result.scalar_one()

    return Pagination(items, page, per_page, total)



from fastapi import FastAPI, Request

from ..config import settings


def create_app() -> FastAPI:
    app = FastAPI(
        title=f'NASBio Web Services - {settings.SERVER_ID}',
        openapi_url=f'/api/{settings.SERVER_ID}/openapi.json',
        docs_url=f'/api/{settings.SERVER_ID}/docs',
        redoc_url=f'/api/{settings.SERVER_ID}/redoc',
    )

    @app.get(f'/api/{settings.SERVER_ID}/')
    async def index():
        return {'message': f'{settings.SERVER_ID.upper()} service is up!'}

    @app.middleware('http')
    async def add_header(request: Request, call_next):
        response = await call_next(request)
        response.headers['server-id'] = settings.SERVER_ID

        return response

    return app



import os
from typing import Generic, Optional, Tuple, TypedDict, TypeVar, Union
from enum import Enum
from datetime import datetime
from dateutil.parser import parse
from base64 import b64encode, b64decode

import json
from fastapi import Depends, HTTPException
import strawberry
from strawberry.extensions import Extension
from strawberry.fastapi.router import Request
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from nasbio import settings
from nasbio.db import get_async_session, get_session

from .sqlalchemy import Pagination

class Group(Enum):
    pass


class Role(Enum):
    user = 'user'
    admin = 'admin'


class Permission(Enum):
    pass


class Authorization(TypedDict):
    groups: list[Group]
    roles: list[Role]
    permissions: list[Permission]


class UserPayloadDict(TypedDict):
    authorization: Authorization
    iss: str
    sub: str
    aud: list[str]
    iat: int
    exp: int


class AuthExtension(Extension):
    def on_request_start(self):
        current_user: UserPayloadDict = self.execution_context.context.get('current_user')
        if (
            # Valid auth credentials.
            (current_user and current_user['sub'])
            or (
            # Development environment.
            (os.environ.get('FASTAPI_ENV') == 'development' or os.environ.get('FASTAPI_ENV') == 'dev_inmemory')
            and (
                # Subgraph introspection by apollo gateway.
                self.execution_context.query == 'query __ApolloGetServiceDefinition__ { _service { sdl } }'
                or
                # Subgraph introspection by rover.
                self.execution_context.query == 'query SubgraphIntrospectQuery {\n    # eslint-disable-next-line'
                                                '\n    _service {\n        sdl\n    }\n}'
                or
                # GraphiQL request.
                self.execution_context.context.get('request').headers.get('referer')
                == f'https://localhost/api/{settings.SERVER_ID}/graphql'
            )
        )
            or
            # Test environment.
            os.environ.get('FASTAPI_ENV') == 'testing'
        ):
            return
        else:
            raise HTTPException(status_code=401, detail='Invalid credentials provided.')


class Context(TypedDict):
    request: Request
    current_user: Optional[UserPayloadDict]
    db: Session
    async_db: AsyncSession


async def get_context(
    request: Request,
    db=Depends(get_session),
    async_db=Depends(get_async_session),
) -> Context:
    current_user_header: str = request.headers.get('current-user')
    user_payload: Optional[UserPayloadDict] = json.loads(current_user_header) \
        if (current_user_header and current_user_header != 'undefined') else None

    return {
        'current_user': user_payload,
        'db': db,
        'async_db': async_db,
    }


# class IsAdmin(BasePermission):
#     message = 'Admin required.'
#
#     def has_permission(self, source: Any, info: Info, **kwargs) -> bool:
#         return info.context['current_user']['admin']


DateTime = strawberry.scalar(
    datetime,
    serialize=lambda value: value.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z',
    parse_value=lambda value: datetime.fromisoformat(value.rstrip('Z')),
)


@strawberry.interface
class MutationResponse:
    success: bool
    message: str


T = TypeVar('T')


@strawberry.type(description='A list of edges.')
class Edge(Generic[T]):
    node: T
    cursor: str


@strawberry.federation.type(description='Information about pagination in a connection.', shareable=True)
class PageInfo:
    start_cursor: Optional[str] = \
        strawberry.field(description='The cursor to continue from when paginating backward.')
    end_cursor: Optional[str] = strawberry.field(description='The cursor to continue from when paginating forward.')
    has_next_page: bool = strawberry.field(description='Whether there are more items when paginating forward.')
    has_previous_page: bool = strawberry.field(description='Whether there are more items when paginating backward.')


@strawberry.type(description='A list of edges with pagination information.')
class Connection(Generic[T]):
    edges: list[Edge[T]]
    page_info: PageInfo
    total_count: int
    filtered_count: int
    page_count: int
    current_page: int

    @classmethod
    def load(cls, data: Pagination, counts: Tuple[int, int]):
        total_count, filtered_count = counts

        return Connection(
            edges=[Edge(node=item, cursor=to_cursor_hash(item.created_at if hasattr(item, 'created_at') else item.id))
                   for item in data.items],
            page_info=PageInfo(
                start_cursor=to_cursor_hash(
                    data.items[0].created_at if hasattr(data.items[0], 'created_at')
                    else data.items[0].id) if data.items else None,
                end_cursor=to_cursor_hash(
                    data.items[len(data.items) - 1].created_at if hasattr(data.items[len(data.items) - 1], 'created_at')
                    else data.items[len(data.items) - 1].id) if data.items else None,
                has_next_page=data.has_next,
                has_previous_page=data.has_prev,
            ),
            total_count=total_count,  # total_count.
            filtered_count=filtered_count,  # filtered_count.
            page_count=data.pages,  # page_count.
            current_page=data.page,  # current_page.
        )


def to_cursor_hash(datetime_or_id: Union[datetime, strawberry.ID]) -> str:
    return str(b64encode(str(datetime_or_id).encode('utf-8')), 'utf-8')


def from_cursor_hash(cursor: str) -> Union[datetime, strawberry.ID]:
    decoded = str(b64decode(cursor), 'utf-8')
    try:
        parse(decoded)
        return datetime.fromisoformat(decoded)
    except:  # noqa
        return decoded



import os
import pathlib
from typing import Optional
from functools import lru_cache
from multiprocessing import cpu_count


class BaseConfig:
    SERVER_ID: str = pathlib.Path(os.getcwd()).parts[-1]
    PROCESS_COUNT: int = cpu_count()

    # Message broker configs.
    MESSAGE_BROKER_URL = \
        f'amqp://{os.environ.get("MESSAGE_BROKER_USER", "guest")}' \
        f':{os.environ.get("MESSAGE_BROKER_PASSWORD", "guest")}' \
        f'@{os.environ.get("MESSAGE_BROKER_HOST", "localhost")}:{os.environ.get("MESSAGE_BROKER_PORT", 5672)}/%2f'

    # SQLAlchemy configs.
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    PG_USER = os.environ.get('PG_USER', 'postgres')
    PG_PASSWORD: Optional[str] = os.environ.get('PG_PASSWORD')
    PG_HOST = os.environ.get('PG_HOST', 'localhost')
    PG_PORT = os.environ.get('PG_PORT', 5432)
    PG_DB = os.environ.get('PG_DB', SERVER_ID if os.environ.get('FASTAPI_ENV') == 'production' else f'{SERVER_ID}_dev')

    @property
    def SQLALCHEMY_DATABASE_URI(self):
        return f'postgresql+psycopg2://{self.PG_USER}:{self.PG_PASSWORD}@{self.PG_HOST}:{self.PG_PORT}/{self.PG_DB}'

    @property
    def ASYNC_SQLALCHEMY_DATABASE_URI(self):
        return f'postgresql+asyncpg://{self.PG_USER}:{self.PG_PASSWORD}@{self.PG_HOST}:{self.PG_PORT}/{self.PG_DB}'

    # Celery configs.
    # broker_url = f'redis://{os.environ.get("REDIS_HOST")}:{os.environ.get("REDIS_PORT", 6379)}/0'
    broker_url = f'amqp://{os.environ.get("RABBITMQ_DEFAULT_USER")}:{os.environ.get("RABBITMQ_DEFAULT_PASS")}' \
                 f'@{os.environ.get("RABBITMQ_HOST")}:{os.environ.get("RABBITMQ_PORT")}'
    result_backend = f'redis://{os.environ.get("REDIS_HOST")}:{os.environ.get("REDIS_PORT", 6379)}/0'
    accept_content = ['application/json']
    task_serializer = 'json'
    result_serializer = 'json'
    redis_max_connections = 5


class DevelopmentConfig(BaseConfig):
    """Development configuration."""
    DEBUG = True


class DevInMemoryConfig(BaseConfig):
    """Development configuration with sqlite in memory."""
    DEBUG = True
    SQLALCHEMY_DATABASE_URI = 'sqlite+pysqlite:///file:rnas?mode=memory&cache=shared&uri=true'
    ASYNC_SQLALCHEMY_DATABASE_URI = 'sqlite+aiosqlite:///file:rnas?mode=memory&cache=shared&uri=true'


class ProductionConfig(BaseConfig):
    """Production configuration."""
    DEBUG = False


class ProdInMemoryConfig(BaseConfig):
    """Production configuration."""
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = 'sqlite+pysqlite:///file:rnas?mode=memory&cache=shared&uri=true'
    ASYNC_SQLALCHEMY_DATABASE_URI = 'sqlite+aiosqlite:///file:rnas?mode=memory&cache=shared&uri=true'


class TestingConfig(BaseConfig):
    """Default testing configuration with sqlite in memory."""
    TESTING = True
    SQLALCHEMY_DATABASE_URI = 'sqlite://'


class TestingConfigPG(BaseConfig):
    """Testing configuration with PostgreSQL."""
    TESTING = True


@lru_cache()
def get_settings():
    config_cls_dict = {
        'development': DevelopmentConfig,
        'dev_inmemory': DevInMemoryConfig,
        'production': ProductionConfig,
        'prod_inmemory': ProdInMemoryConfig,
        # Use PostgreSQL DB if USE_PG environment variable is set, else default to inmemory sqlite for testing.
        'testing': TestingConfigPG if os.environ.get('USE_PG', '') else TestingConfig,
    }

    config_name = os.environ.get('FASTAPI_ENV', 'development')
    config_cls = config_cls_dict[config_name]

    return config_cls()


settings = get_settings()



from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base

meta = MetaData(naming_convention={
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
})
Base = declarative_base(metadata=meta)

# from typing import Any
# from sqlalchemy import inspect
# from sqlalchemy.ext.declarative import as_declarative, declared_attr
#
# @as_declarative()
# class Base:
#     id: Any
#     __name__: str
#
#     # Generate __tablename__ automatically
#     @declared_attr
#     def __tablename__(cls) -> str:
#         return cls.__name__.lower()
#
#     def to_dict(self):
#         return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nasbio', 'nasbio.app', 'nasbio.db', 'nasbio.events']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.87.0,<0.88.0',
 'pika>=1.3.1,<2.0.0',
 'sqlalchemy>=1.4.43,<2.0.0',
 'strawberry-graphql[fastapi]>=0.142.2,<0.143.0',
 'uvicorn[standard]>=0.19.0,<0.20.0']

setup_kwargs = {
    'name': 'nasbio',
    'version': '0.1.24',
    'description': '',
    'long_description': '# NASBio Web Services\n## Common Python Packages\n',
    'author': 'Kah Wai LIM',
    'author_email': 'kahwai222@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
